const _S           = String;


const raw          = _S.raw;
const cd2ch        = (c)=>_S.fromCodePoint(c); 
const ch2cd        = (s)=>s.codePointAt(0);

const ascii_cd2ch  = (c)=>_S.fromCharCode(c);
const ascii_ch2cd  = (s)=>s.charCodeAt(0);

const is_high_surrogate  =(n)=>(n>>8 & 0b1111_1100) === 0b1101_1000;
const is_low_surrogate   =(n)=>(n>>8 & 0b1111_1100) === 0b1101_1100;
const is_not_surrogate   =(n)=>(!is_high_surrogate(n)) && (!is_low_surrogate(n));


const surrogate_to_code     = (n16_0,n16_1)=> (
     ((n16_0 & 0x03ff) << 10) +
     ((n16_1 & 0x03ff)) +
     0x10000
);


const chlen = (s) => {
    let ei  = s.length;
    let i   = 0;
    let cnt = 0;
    let cd;
    while(i<ei) {
        cd = s[i].charCodeAt(0);
        ++cnt;
        if((cd & 0xfc00) === 0xd800) {
            ++i;
            ++i;
        } else {
            ++i;
        }
    }
    return(cnt)
}

const cdat   = (s,chidx) => {
    let ei   = s.length;
    let i    = 0;
    let idx  = 0;
    let code = -1;
    let cd;
    while(i<ei) {
        cd = ascii_ch2cd(s[i]);
        if(is_high_surrogate(cd)) {
            code = surrogate_to_code(cd,ascii_ch2cd(s[i+1]));
            if(idx === chidx) {
                return(code)
            } else {
                ++i;
                ++i;
                ++idx;
            }
        } else {
            code = cd;
            if(idx === chidx) {
                return(code)
            } else {
                ++i;
                ++idx;
            }
        }
    }
}

const chat = (s,chidx) => cd2ch(cdat(s,chidx));

const slice_by_chidx = (s,si,ei) => {
    let i =0;
    let idx =0;
    let cd;
    let ssi;
    let sei;
    while(i<s.length) {
        if(idx === si) {ssi = i}
        if(idx === ei) {
            sei = i;
            break;
        }
        cd = ascii_ch2cd(s[i]);
        if(is_high_surrogate(cd)) {
            ++i;++i;
            ++idx;
        } else {
            ++i;
            ++idx;
        }
    }
    return(s.slice(ssi,sei))
}

const _avgcut_by_chlen = (s,sz) => {
    let arr = []
    let i =0;
    let cd;
    let pair = [0,-1];
    let cnt  = 0;
    while(i<s.length) {
        if(cnt === sz) {
            arr.push([pair[0],pair[1]])
            pair[0] = pair[1];
            pair[1] = -1;
            cnt =0;
        } 
        cd = ascii_ch2cd(s[i]);
        if(is_high_surrogate(cd)) {
            ++i;++i;
        } else {
            ++i;
        }
        ++cnt;
        pair[1] = i;
    }
    if(pair[0] === s.length) {
    } else {
        pair[1] = s.length;
        arr.push(pair)
    }
    return(arr)
}

const avgcut_by_chlen = (s,sz) => {
    let arr = []
    let i =0;
    let cd;
    let pair = [0,-1];
    let cnt  = 0;
    while(i<s.length) {
        if(cnt === sz) {
            arr.push(s.slice(pair[0],pair[1]))
            pair[0] = pair[1];
            pair[1] = -1;
            cnt =0;
        }
        cd = ascii_ch2cd(s[i]);
        if(is_high_surrogate(cd)) {
            ++i;++i;
        } else {
            ++i;
        }
        ++cnt;
        pair[1] = i;
    }
    if(pair[0] === s.length) {
    } else {
        pair[1] = s.length;
        arr.push(s.slice(pair[0],pair[1]))
    }
    return(arr)
}

const ucs2_ablen = (s)=> s.length * 2;

const _avgcut_by_ucs2_ablen = (s,sz) => {
    sz = (sz - sz%2);
    if(sz <4) {throw("sz_must_be_even_and_ge_4")} else {}
    ///------------------------------------------------------------
    let len = s.length;
    let arr = []
    let i   = 0;
    let pair = [0,-1];
    let cnt  = 0;
    while(i<len) {
        cd = ascii_ch2cd(s[i]);
        let cond = is_high_surrogate(cd);
        if(cond) {
            if(
                cnt === sz || (cnt=== (sz-2))
            ) {
                arr.push([pair[0],pair[1]])
                pair[0] = pair[1];
                pair[1] = -1;
                cnt =0;
            }
            ++i;++i;
            cnt = cnt+4;
        } else {
            if(
                cnt === sz
            ) {
                arr.push([pair[0],pair[1]])
                pair[0] = pair[1];
                pair[1] = -1;
                cnt =0;
            }
            ++i;
            cnt = cnt+2;
        }
        pair[1] = i;
    }
    if(pair[0] === s.length) {
    } else {
        pair[1] = s.length;
        arr.push(pair)
    }
    return(arr)
}

const avgcut_by_ucs2_ablen = (s,sz) => {
    sz = (sz - sz%2);
    if(sz <4) {throw("sz_must_be_even_and_ge_4")} else {}
    ///------------------------------------------------------------
    let len = s.length;
    let arr = []
    let i   = 0;
    let pair = [0,-1];
    let cnt  = 0;
    while(i<len) {
        cd = ascii_ch2cd(s[i]);
        let cond = is_high_surrogate(cd);
        if(cond) {
            if(
                cnt === sz || (cnt=== (sz-2)) 
            ) {
                arr.push(s.slice(pair[0],pair[1]))
                pair[0] = pair[1];
                pair[1] = -1;
                cnt =0;
            }
            ++i;++i;
            cnt = cnt+4;
        } else {
            if(
                cnt === sz
            ) {
                arr.push(s.slice(pair[0],pair[1]))
                pair[0] = pair[1];
                pair[1] = -1;
                cnt =0;
            }
            ++i;
            cnt = cnt+2;
        }
        pair[1] = i;
    }
    if(pair[0] === s.length) {
    } else {
        pair[1] = s.length;
        arr.push(s.slice(pair[0],pair[1]))
    }
    return(arr)
}


////-------------------utf8-----------
const _utf8 = require("./utf8");

const utf8_ablen = (s) => {
    let ei  = s.length;
    let i   = 0;
    let cnt = 0;
    let cd;
    while(i<ei) {
        cd = s[i].charCodeAt(0);
        if((cd & 0xfc00) === 0xd800) {
            cd = surrogate_to_code(cd,ascii_ch2cd(s[i+1]));
            cnt = cnt + _utf8.len(cd);
            ++i;
            ++i;
        } else {
            cnt = cnt + _utf8.len(cd);
            ++i;
        }
    }
    return(cnt)        
}


const roughly_split_by_utf8_ablen = (s,sz) => {
    sz = (sz - sz%4)/4;
    sz = (sz < 32)?32:sz;
    let arr =[];
    let si = 0;
    while(true) {
        ei = si + sz;
        if(ei>=s.length) {
            arr.push(s.slice(si,s.length));
            break;
        } else {
            let cd = s[ei].charCodeAt(0);
            if(is_low_surrogate(cd)) {
                ei = ei -1;
            } else {
            }
            arr.push(s.slice(si,ei));
            si = ei;
        }
    }
    return(arr)
}



module.exports = {
   raw, 
   ch2cd,cd2ch, 
   ascii_cd2ch,ascii_ch2cd, 
   is_high_surrogate,
   is_low_surrogate,
   is_not_surrogate,
   surrogate_to_code,
   chlen,
   cdat, 
   chat, 
   ////
   slice_by_chidx,
   _avgcut_by_chlen,
   avgcut_by_chlen,
   ////
   ucs2_ablen,
   _avgcut_by_ucs2_ablen, 
   avgcut_by_ucs2_ablen, 
   //// 
   _utf8,
   utf8_ablen,
   roughly_split_by_utf8_ablen, 
}

